﻿#version 400 core
out vec4 FragColor;

in vec2 TexCoords;

// Sun information
uniform vec3 viewPos;
uniform vec3 lightDir;
uniform vec3 lightColor;

// Ambient information
uniform vec3 ambientColor;
uniform float ambientIntensity;

// Shadow settings
uniform mat4 shadowMatrix;
uniform int usePixelShadows;
uniform int pixelShadowScale;
uniform float shadowBiasMax;
uniform float shadowBiasMin;

// GBuffer
uniform sampler2D gPosition;
uniform sampler2D gNormal;
uniform sampler2D gAlbedoSpec;

// Other textures
uniform sampler2D shadowMap;

// Position recreation
uniform sampler2D depthBuffer;

// Pixel Shadows
uniform mat4 pixelShadowTransformation;
uniform mat4 inversePixelShadowTransformation;

uniform mat4 projMatrixInv;
uniform mat4 viewMatrixInv;

float when_gt(float x, float y) {
  return max(sign(x - y), 0.0);
}

float ShadowCalculation(vec4 fragLightSpace, vec3 normal, vec3 lightDir)
{
    // Convert frag light space to texel coordinates
    vec3 projCoords = fragLightSpace.xyz / fragLightSpace.w;
    projCoords = projCoords * 0.5 + 0.5;

    float closestDepth = texture(shadowMap, projCoords.xy).r;

    float currentDepth = projCoords.z;

    float shadow = 0;
    float bias = max(shadowBiasMax * (1.0 - dot(normal, lightDir)), shadowBiasMin);
    if(usePixelShadows == 1)
    {
        shadow = currentDepth >= closestDepth ? 1.0 : 0.0;

        if(shadow == 0.0)
        {
            vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
            for(int x = -1; x <= 1; ++x)
            {
                for(int y = -1; y <= 1; ++y)
                {
                    float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
                    shadow += (currentDepth - bias >= pcfDepth ? 1.0 : 0.0);        
                }
            }
            shadow /= 9.0;
        }
    }
    else
    {
        vec2 texelSize = 0.5 / textureSize(shadowMap, 0);
        for(int x = -1; x <= 1; ++x)
        {
            for(int y = -1; y <= 1; ++y)
            {
                float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
                shadow += (currentDepth - bias >= pcfDepth ? 1.0 : 0.0);        
            }
        }
        shadow /= 9.0;
    }

    if(projCoords.z > 1.0)
        shadow = 0.0;

    return shadow;
}

// this is supposed to get the world position from the depth buffer
vec3 WorldPosFromDepth(float depth) {
    float z = depth * 2.0 - 1.0;

    vec4 clipSpacePosition = vec4(TexCoords * 2.0 - 1.0, z, 1.0);
    vec4 viewSpacePosition = clipSpacePosition * projMatrixInv;

    // Perspective division
    viewSpacePosition /= viewSpacePosition.w;

    vec4 worldSpacePosition = viewSpacePosition * viewMatrixInv;

    return worldSpacePosition.xyz;
}

vec3 ViewPosFromDepth(float depth)
{
    float z = depth * 2.0 - 1.0;

    vec4 clipSpacePosition = vec4(TexCoords * 2.0 - 1.0, z, 1.0);
    vec4 viewSpacePosition = clipSpacePosition * projMatrixInv;

    // Perspective division
    return (viewSpacePosition / viewSpacePosition.w).xyz;
}

void main()
{             
    // retrieve data from gbuffer
    float depth = texture(depthBuffer, TexCoords).r;
    vec3 Normal = texture(gNormal, TexCoords).rgb;
    vec3 Albedo = texture(gAlbedoSpec, TexCoords).rgb;
    float Specular = 0;

    vec3 shadowPos = ViewPosFromDepth(depth);
    if(usePixelShadows == 1)
    {
        vec3 magicPos = (vec4(shadowPos, 1.0) * pixelShadowTransformation).xyz;
        double magicX = double(magicPos.x);
        double magicY = double(magicPos.y);
        double magicZ = double(magicPos.z);

        magicX = round(magicX * pixelShadowScale) / pixelShadowScale;
        magicY = round(magicY * pixelShadowScale) / pixelShadowScale;
        magicZ = round(magicZ * pixelShadowScale) / pixelShadowScale;

        magicPos = vec3(float(magicX), float(magicY), float(magicZ));//round(magicPos * pixelShadowScale) / pixelShadowScale;
        shadowPos = (vec4(magicPos, 1.0) * inversePixelShadowTransformation * viewMatrixInv).xyz;
    }
    else
    {
        shadowPos = (vec4(shadowPos, 1.0) * viewMatrixInv).xyz;
    }

    // Shadow calculation
    float shadow = ShadowCalculation(vec4(shadowPos, 1.0) * shadowMatrix, Normal, lightDir);

    // Base lighting, 10% if no other light
    vec3 ambient = ambientColor * ambientIntensity;

    // Diffuse
    float diffuse = max(dot(Normal, lightDir), 0.0);

    // Specular
    //vec3 viewDir = normalize(viewPos - FragPos);
    //vec3 reflectDir = reflect(-lightDir, Normal);
    //vec3 halfwayDir = normalize(lightDir + viewDir);

    float specular = 0;//pow(max(dot(Normal, halfwayDir), 0.0), 256 * Specular) * when_gt(Specular, 0.0) * when_gt(diffuse, 0.0);

    // Final result
    FragColor = vec4(Albedo * (1.0 - shadow) * (specular + diffuse) * lightColor + Albedo * ambient, 1);

    // Normal Test
    //FragColor = vec4((Normal + 1) / 2, 1.0);

    // Position Test
    //FragColor = vec4(FragPos / 10, 1.0);

    // Shadow Test
    //FragColor += vec4(vec3(1 - texture(shadowMap, TexCoords).r), 1.0);

    // Depth test
    //FragColor = vec4(depth, depth, depth, 1.0);

    gl_FragDepth = depth;
}